#include "FileUtf16in.h"



/** FUNZIONI MEMBRO --- INIZIALIZZAZIONE **/

/* Apre un file:
 *  - in caso di errore restituisce false.
 */
void FileUtf16in::open(const char filename[]) {
    file=fopen(filename, "rb");
    if(file==NULL) throw Errore(getFileStat(), ERR_FO_OPEN_FAIL);
}



/* Chiude il file aperto:
 */
void FileUtf16in::close(void) {
    fclose(file);
}




/** FUNZIONI MEMBRO --- LETTURA DATI **/

/* Legge un carattere:
 *  - scarta il byte di ordine inferiore;
 *  - se si passa NULL legge un carattere e lo scarta (=> avanza di due bytes nel file)
 *  - in caso di errore di lettura salva il carattere nullo.
 */
void FileUtf16in::readChar(char *ch) {
    u8 duebytes[2];
    u16 bytesLetti;
    //Legge due byte consecutivi
    bytesLetti=fread(duebytes, 1, 2, file); //Restituisce il numero di oggetti(=1B) letti
    if(ch!=NULL) {
        if(bytesLetti==2) {
            //Ha letto correttamente 2 bytes => salva il byte di ordine maggiore (come se fosse testo ascii)
            *ch=duebytes[0];
        }
        else {
            //Raggiunta la fine del file o altro errore di lettura (es. file chiuso)
            throw Errore(getFileStat(), ERR_FO_READ_CHAR);
            //*ch='\0';
        }
    }
    return;
}


/* Legge una stringa:
 *  - svuota la stringa fornita in ingresso se contiene uno o pi caratteri;
 *  - scarta il byte di ordine inferiore di ogni carattere;
 *  - la stringa da leggere deve essere tutta contenuta sulla riga corrente e non deve contenere
 *    i caratteri 'non alfanumerici' descritti nella funzione 'bool isStrTerm(ch)';
 *  - i caratteri terminatori non sono contenuti nella stringa ma vengono gi letti.
 */
void FileUtf16in::readString(string *str) {
    try {
        char ch;
        //Svuota la stringa se contiene qualcosa
        if(!str->empty()) {
            str->erase();
        }
        //Legge il primo carattere
        readChar(&ch);
        //Salva i caratteri nella stringa fin quando raggiunge il terminatore
        while(!isStrTerm(ch)) {
            *str+=ch;
            readChar(&ch);
        }
    }
    catch(Errore err) {
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_READ_STR);
        throw err;
    }
}


/* Legge un intero:
 *  - senza segno;
 *  - lunghezza 32bit;
 *  - i caratteri terminatori vengono gi letti;
 *  - lancia errore nel caso vengano oltrepassati i limiti o di lettura carattere non valido.
 */
void FileUtf16in::readInt(u32 *num) {
    try {
        char ch, segno;
        string numero;
        do {
            //Legge il primo carattere, saltando segni di formattazione
            readChar(&ch);
        } while(isStrTerm(ch));
        //Controlla cosa ha letto
        if(isSign(ch)) {
            //E' un carattere di segno
            segno=ch;
        }
        else  {
            //Se non  un carattere di segno, controlla se  un numero
            if(isNum(ch)) {
                //E' un numero, allora lo aggiunge alla stringa e mette segno positivo
                numero=ch;
                segno='+';
            }
            else {
                //E' un carattere non valido
                throw Errore(ERR_OP_INT_READ, ERR_TP_UNEXPECTED_CHAR);
            }
        }
        //Legge il secondo carattere
        readChar(&ch);
        //Salva i caratteri nella stringa finch sono numeri (eof implica ch=='\0' => isNum==false)
        while(isNum(ch)) {
            numero+=ch;
            readChar(&ch);
        }
        //Controlla la lunghezza
        checkIntLenght(numero, +1);
        //Converte la stringa in numero
        string2int(numero, '+', num);
    }
    catch(Errore err) {
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_READ_UINT);
        throw err;
    }
}


/* Legge un intero:
 *  - con segno;
 *  - lunghezza 32bit;
 *  - i caratteri terminatori vengono gi letti;
 *  - lancia errore nel caso vengano oltrepassati i limiti o di lettura carattere non valido.
 */
void FileUtf16in::readInt(s32 *num) {
    try {
        char ch, segno;
        string numero;
        do {
            //Legge il primo carattere, saltando segni di formattazione
            readChar(&ch);
        } while(isStrTerm(ch));
        //Controlla cosa ha letto
        if(isSign(ch)) {
            //E' un carattere di segno
            segno=ch;
        }
        else  {
            //Se non  un carattere di segno, controlla se  un numero
            if(isNum(ch)) {
                //E' un numero, allora lo aggiunge alla stringa e mette segno positivo
                numero=ch;
                segno='+';
            }
            else {
                //E' un carattere non valido
                throw Errore(ERR_OP_INT_READ, ERR_TP_UNEXPECTED_CHAR);
            }
        }
        //Legge il secondo carattere
        readChar(&ch);
        //Salva i caratteri nella stringa finch sono numeri (eof implica ch=='\0' => isNum==false)
        while(isNum(ch)) {
            numero+=ch;
            readChar(&ch);
        }
        //Controlla la lunghezza
        checkIntLenght(numero, -1);
        //Converte la stringa in numero
        string2int(numero, segno, num);
    }
    catch(Errore err) {
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_READ_SINT);
        throw err;
    }
}


/* Legge un intero:
 *  - senza segno;
 *  - lunghezza 16bit;
 *  - i caratteri terminatori vengono gi letti;
 *  - lancia errore nel caso vengano oltrepassati i limiti o di lettura carattere non valido.
 */
 void FileUtf16in::readInt(u16 *num) {
    u32 letto;
    readInt(&letto);
    if(letto>65535) {
        //Va oltre i limiti dell'intero u16
        throw Errore(ERR_OP_INT_CONV, ERR_TP_EXCEED_UNITS);
    }
    *num=(u16)letto;
}


/* Legge un intero:
 *  - senza segno;
 *  - lunghezza 8bit;
 *  - i caratteri terminatori vengono gi letti;
 *  - lancia errore nel caso vengano oltrepassati i limiti o di lettura carattere non valido.
 */
 void FileUtf16in::readInt(u8 *num) {
    u32 letto;
    readInt(&letto);
    if(letto>255) {
        //Va oltre i limiti dell'intero u8
        throw Errore(ERR_OP_INT_CONV, ERR_TP_EXCEED_UNITS);
    }
    *num=(u8)letto;
}


/* Legge un intero:
 *  - con segno;
 *  - lunghezza 16bit;
 *  - i caratteri terminatori vengono gi letti;
 *  - lancia errore nel caso vengano oltrepassati i limiti o di lettura carattere non valido.
 */
 void FileUtf16in::readInt(s16 *num) {
    s32 letto;
    readInt(&letto);
    if(letto<-32767 || letto>32767) {
        //Va oltre i limiti dell'intero s16
        throw Errore(ERR_OP_INT_CONV, ERR_TP_EXCEED_UNITS);
    }
    *num=(u16)letto;
}


/* Legge un intero:
 *  - con segno;
 *  - lunghezza 8bit;
 *  - i caratteri terminatori vengono gi letti;
 *  - lancia errore nel caso vengano oltrepassati i limiti o di lettura carattere non valido.
 */
 void FileUtf16in::readInt(s8 *num) {
    s32 letto;
    readInt(&letto);
    if(letto<-127 || letto>127) {
        //Va oltre i limiti dell'intero s8
        throw Errore(ERR_OP_INT_CONV, ERR_TP_EXCEED_UNITS);
    }
    *num=(u8)letto;
}


/* Legge un float:
 *  - con/senza segno;
 *  - al max 6 cifre di precisione (float 32bit);
 *  - riconosce , e . come 'virgola';
 *  - riconosce e o E come 10^;
 *  - i caratteri terminatori vengono gi letti;
 *  - salva 0 in caso di errore.
 */
void FileUtf16in::readFloat(f32 *num) {
    try {
        char ch, segno='\0', segnoMant='+', segnoExp='+';
        string mantInt, mantDec, exp;
        bool go=true, decFlag=false, expFlag=false, signUsed=true;

        //Riempie le 3 stringhe che rappresentano il numero
        do {
            readChar(&ch);
            if(isSign(ch)) {
                //E' un carattere di segno, ma non so se  della mantissa o dell'esponente
                segno=ch;
                signUsed=false;
            }
            else {
                if(isDot(ch)) {
                    //Ha trovato la virgola
                    decFlag=true;
                    expFlag=false;
                }
                else if(isExp(ch)) {
                    //Ha trovato il segno di esponente
                    decFlag=false;
                    expFlag=true;
                }
                else if(isNum(ch)) {
                    //Ha letto un numero => lo assegna a una parte del numero
                    if(decFlag==true) {
                        //Deve memorizzare la parte decimale della mantissa
                        mantDec+=ch;
                        if(signUsed==false) {
                            //Se non ha assegnato ancora il segno letto,  della mantissa
                            segnoMant=segno;
                            signUsed=true;
                        }
                    }
                    else if(expFlag==true) {
                        //Deve memorizzare l'esponente
                        exp+=ch;
                        if(signUsed==false) {
                            //Se non ha assegnato ancora il segno letto,  dell'esponente
                            segnoExp=segno;
                            signUsed=true;
                        }
                    }
                    else {
                        //Deve memorizzare la parte intera della mantissa
                        mantInt+=ch;
                        if(signUsed==false) {
                            //Se non ha assegnato ancora il segno letto,  della mantissa
                            segnoMant=segno;
                            signUsed=true;
                        }
                    }
                }
                else {
                    //Non  neppure un numero => termina la lettura
                    go=false;
                }
            }
        } while(go);
        //Converte la stringa in numero
        string2float(mantInt, mantDec, exp, segnoMant, segnoExp, num);
    }
    catch(Errore err) {
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_READ_FLOAT);
        throw err;
    }
}




/** FUNZIONI MEMBRO --- SCARTA DATI LETTI **/

/* Scarta n campi di dati:
 *  - i caratteri terminatori vengono gi letti;
 */
void FileUtf16in::skip(u8 n) {
    try {
        char ch;
        do {
            //Legge il primo carattere, saltando segni di formattazione
            readChar(&ch);
        } while(isStrTerm(ch));
        for(u8 i=0; i<n; i++) {
            do {
                readChar(&ch);
            } while(!isStrTerm(ch) && ch!='\0');
        }
    }
    catch(Errore err) {
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_SKIP_BYTES);
        throw err;
    }
}




/** FUNZIONI MEMBRO --- RICERCA STRINGHE **/

/* Cerca una stringa:
 *  - se la trova restituisce 1;
 *  - altrimenti restituisce 0;
 *  - in caso di errore di lettura lancia errore (anche eof => nessun controllo sul ciclo)
 */
u8 FileUtf16in::findString(char str[]) {
    try {
        char ch;
        u16 lung, i=0;

        lung=strlen(str)-1;
        do {
            readChar(&ch);
            if(ch==str[i]) {
                //Il carattere letto  uguale all'i-esimo di str
                if(i==lung) {
                    //Se i=lung ha finito (NB: lung  gi stato ridotto di 1 e i parte da 0)
                    return 1;
                }
                else {
                    //Se i<lung aumenta il contatore
                    i++;
                }
            }
            else {
                //Il carattere letto  diverso dall'i-esimo di str => azzera il contatore
                i=0;
                if(!isForm(ch)) {
                    //Se l'ultimo carattere letto non  di formattazione (es sp, tab) termina la riga
                    while(ch!='\n') readChar(&ch);
                }
            }
        } while(true);
        //Se arriva a questo punto non ha trovato la stringa
        return 0;
    }
    catch(Errore err) {
        //Se l'errore  eof, restituisce zero
        if(err.getFileStat()==FILE_ST_EOF) return 0;
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_1STR_SEARCH);
        throw err;
    }
}


/* Cerca una stringa:
 *  - se trova la prima restituisce 1;
 *  - se trova la seconda restituisce 2;
 *  - altrimenti restituisce 0;
 *  - in caso di errore di lettura lancia errore (anche eof => nessun controllo sul ciclo)
 */
u8 FileUtf16in::findString(char str1[], char str2[]) {
    try {
        char ch;
        u16 lung1, lung2, i=0, j=0;

        lung1=strlen(str1)-1;
        lung2=strlen(str2)-1;
        do {
            readChar(&ch);
            //Confronto con str1
            if(ch==str1[i] || ch==str2[j]) {
                //Il carattere  presente in una delle due stringhe
                //Controlla la prima
                if(ch==str1[i]) {
                    //Il carattere letto  uguale all'i-esimo di str1
                    if(i==lung1) {
                        //Se i=lung1 ha finito (NB: lung  gi stato ridotto di 1 e i parte da 0)
                        return 1;
                    }
                    else {
                        //Se i<lung aumenta il contatore
                        i++;
                    }
                }
                else {
                    //Il carattere letto  diverso dall'i-esimo di str1 => azzera il contatore
                    i=0;
                }
                //Controlla la seconda
                if(ch==str2[j]) {
                    //Il carattere letto  uguale all'i-esimo di str1
                    if(j==lung2) {
                        //Se j=lung2 ha finito (NB: lung  gi stato ridotto di 1 e j parte da 0)
                        return 2;
                    }
                    else {
                        //Se i<lung aumenta il contatore
                        j++;
                    }
                }
                else {
                    //Il carattere letto  diverso dal j-esimo di str2 => azzera il contatore
                    j=0;
                }
            }
            else {
                //Il carattere non appartiene a nessuna stringa => azzera i contatori
                i=0;
                j=0;
                if(!isForm(ch)) {
                    //Se l'ultimo carattere letto non  di formattazione (es sp, tab, endl) termina la riga
                    while(ch!='\n') readChar(&ch);
                }
            }
        } while(true);
        //Se arriva a questo punto non ha trovato alcuna stringa
        return 0;
    }
    catch(Errore err) {
        //Se l'errore  eof, restituisce zero
        if(err.getFileStat()==FILE_ST_EOF) return 0;
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_2STR_SEARCH);
        throw err;
    }
}


/* Cerca una stringa:
 *  - se trova la prima restituisce 1;
 *  - se trova la seconda restituisce 2;
 *  - se trova la terza restituisce 3;
 *  - altrimenti restituisce 0;
 *  - in caso di errore di lettura lancia errore (anche eof => nessun controllo sul ciclo)
 */
u8 FileUtf16in::findString(char str1[], char str2[], char str3[]) {
    try {
        char ch;
        u16 lung1, lung2, lung3, i=0, j=0, k=0;

        lung1=strlen(str1)-1;
        lung2=strlen(str2)-1;
        lung3=strlen(str3)-1;
        do {
            readChar(&ch);
            //Confronto con str1
            if(ch==str1[i] || ch==str2[j] || ch==str3[k]) {
                //Il carattere  presente in una delle tre stringhe
                //Controlla la prima
                if(ch==str1[i]) {
                    //Il carattere letto  uguale all'i-esimo di str1
                    if(i==lung1) {
                        //Se i=lung1 ha finito (NB: lung  gi stato ridotto di 1 e i parte da 0)
                        return 1;
                    }
                    else {
                        //Se i<lung1 aumenta il contatore
                        i++;
                    }
                }
                else {
                    //Il carattere letto  diverso dall'i-esimo di str1 => azzera il contatore
                    i=0;
                }
                //Controlla la seconda
                if(ch==str2[j]) {
                    //Il carattere letto  uguale al j-esimo di str2
                    if(j==lung2) {
                        //Se j=lung2 ha finito (NB: lung  gi stato ridotto di 1 e j parte da 0)
                        return 2;
                    }
                    else {
                        //Se i<lung2 aumenta il contatore
                        j++;
                    }
                }
                else {
                    //Il carattere letto  diverso dal j-esimo di str2 => azzera il contatore
                    j=0;
                }
                //Controlla la terza
                if(ch==str3[k]) {
                    //Il carattere letto  uguale al k-esimo di str3
                    if(k==lung3) {
                        //Se k=lung3 ha finito (NB: lung  gi stato ridotto di 1 e k parte da 0)
                        return 3;
                    }
                    else {
                        //Se k<lung3 aumenta il contatore
                        k++;
                    }
                }
                else {
                    //Il carattere letto  diverso dal k-esimo di str3 => azzera il contatore
                    k=0;
                }
            }
            else {
                //Il carattere non appartiene a nessuna stringa => azzera i contatori
                i=0;
                j=0;
                k=0;
                if(!isForm(ch)) {
                    //Se l'ultimo carattere letto non  di formattazione (es sp, tab, endl) termina la riga
                    while(ch!='\n') readChar(&ch);
                }
            }
        } while(true);
        //Se arriva a questo punto non ha trovato alcuna stringa
        return 0;
    }
    catch(Errore err) {
        //Se l'errore  eof, restituisce zero
        if(err.getFileStat()==FILE_ST_EOF) return 0;
        //Aggiunge l'informazione relativa all'operazione sul file e rilancia l'errore
        err.setErrFileOp(ERR_FO_3STR_SEARCH);
        throw err;
    }
}




/** FUNZIONI MEMBRO --- STATO **/

/* Reastituisce lo stato del file:
 *  - funzione PRIVATA.
 */
FileStat FileUtf16in::getFileStat(void) {
    if(file==NULL) {
        //File chiuso
        return FILE_ST_CLOSED;
    }
    else {
        //File aperto
        if(feof(file)) {
            //File aperto ma  stata raggiunta la fine
            return FILE_ST_EOF;
        }
        else {
            //File aperto e utilizzabile
            return FILE_ST_OPENED;
        }
    }
}
